import pathlib as _pathlib
from importlib import metadata as _metadata

import toml as _toml

from aok._operations import to_comparator  # noqa
from aok._types import ArbitraryDict  # noqa: F401
from aok._types import ArbitraryList  # noqa: F401
from aok._types import OkayRoot  # noqa: F401
from aok.comparisons import *  # noqa
from aok.comparisons import Anything  # noqa: F401
from aok.comparisons import Between  # noqa: F401
from aok.comparisons import Dict  # noqa: F401
from aok.comparisons import Equals  # noqa: F401
from aok.comparisons import Greater  # noqa: F401
from aok.comparisons import GreaterOrEqual  # noqa: F401
from aok.comparisons import JsonList  # noqa: F401
from aok.comparisons import Less  # noqa: F401
from aok.comparisons import LessOrEqual  # noqa: F401
from aok.comparisons import Like  # noqa: F401
from aok.comparisons import LikeCase  # noqa: F401
from aok.comparisons import List  # noqa: F401
from aok.comparisons import Match  # noqa: F401
from aok.comparisons import NoneOf  # noqa: F401
from aok.comparisons import NotNull  # noqa: F401
from aok.comparisons import Okay  # noqa: F401
from aok.comparisons import OkayList  # noqa: F401
from aok.comparisons import OneOf  # noqa: F401
from aok.comparisons import Optional  # noqa: F401
from aok.comparisons import Tuple  # noqa: F401
from aok.comparisons import Unequals  # noqa: F401
from aok.comparisons import anything  # noqa: F401
from aok.comparisons import between  # noqa: F401
from aok.comparisons import equals  # noqa: F401
from aok.comparisons import greater  # noqa: F401
from aok.comparisons import greater_or_equal  # noqa: F401
from aok.comparisons import json_list  # noqa: F401
from aok.comparisons import less  # noqa: F401
from aok.comparisons import less_or_equal  # noqa: F401
from aok.comparisons import like  # noqa: F401
from aok.comparisons import like_case  # noqa: F401
from aok.comparisons import match  # noqa: F401
from aok.comparisons import none_of  # noqa: F401
from aok.comparisons import not_null  # noqa: F401
from aok.comparisons import one_of  # noqa: F401
from aok.comparisons import optional  # noqa: F401
from aok.comparisons import unequals  # noqa: F401

try:
    __version__ = _metadata.version(__package__)
except _metadata.PackageNotFoundError:  # pragma: no-cover
    # If the package is not installed such that it has distribution metadata
    # fallback to loading the version from the pyproject.toml file.
    __version__ = _toml.loads(
        _pathlib.Path(__file__).parent.parent.joinpath("pyproject.toml").read_text()
    )["tool"]["poetry"]["version"]



import typing

from aok import _definitions

ArbitraryDict = typing.Dict[typing.Any, typing.Any]
ArbitraryList = typing.List[typing.Any]


class OkayRoot(typing.Protocol):  # pragma: no cover
    """Structural-duck-type for root Okay objects."""

    def assert_subset(
        self,
        observed: typing.Union[ArbitraryDict, ArbitraryList],
        message: str = None,
    ):
        """
        Compares the observed object against the expected values in a recursive,
        element-wise fashion and raises assertion errors for any deviation between
        the elements of the expected configuration and the observed structure values.

        In this subset mode, any extra keys/values found in dictionaries will be
        ignored and assumed to be insignificant. Use `assert_all` for exact matching.

        :param observed:
            Data structure to compare against the expected one.
        :param message:
            Optional assertion message to include when differences are found between
            the observed and expected data structures.
        """
        pass

    def assert_all(
        self,
        observed: typing.Union[ArbitraryDict, ArbitraryList],
        message: str = None,
    ):
        """
        Compares the observed object against the expected values in a recursive,
        element-wise fashion and raises assertion errors for any deviation between
        the elements of the expected configuration and the observed structure values.

        :param observed:
            Data structure to compare against the expected one.
        :param message:
            Optional assertion message to include when differences are found between
            the observed and expected data structures.
        """
        pass

    def compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> "_definitions.Comparison":
        """
        Compares the observed object against the expected values in a recursive,
        element-wise fashion and returns a comparison object that expresses the
        results of the compare operation.

        :param observed:
            Data structure to compare against the expected one.
        :param subset:
            When true, any extra keys/values found in dictionaries will be ignored
            and assumed to be insignificant. Set to false for exact matching.
        :return:
            A comparison object that describes the element-wise differences between
            the observed data structure and its expectations.
        """
        pass



import typing

import yaml

from aok import _definitions


class Optional(_definitions.Comparator):
    """Compares two value as an equality or allow None."""

    def __init__(self):
        super(Optional, self).__init__(None)

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Make an optional equality comparison."""
        return observed is None or observed == self.value

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Optional":
        return cls()


class NotNull(_definitions.Comparator):
    """Won't allow a value to be null."""

    def __init__(self):
        super(NotNull, self).__init__(None)

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Make an optional equality comparison."""
        return observed is not None

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "NotNull":
        return cls()


NotNull.register()
not_null = getattr(NotNull, "constructor", NotNull)

Optional.register()
optional = getattr(Optional, "constructor", Optional)

__all__ = [
    "NotNull",
    "not_null",
    "Optional",
    "optional",
]



import typing

import yaml

from aok import _definitions
from aok import _operations


class Equals(_definitions.Comparator):
    """Compares two values as an equality."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Make an equals comparison."""
        return _operations.cast_compatible(self.value, observed) == observed


class Unequals(_definitions.Comparator):
    """Compares two values as an inequality."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Make an inequality comparison."""
        return _operations.cast_compatible(self.value, observed) != observed


class Anything(_definitions.Comparator):
    """Allows anything for the given value."""

    def __init__(self):
        super(Anything, self).__init__(None)

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Anything will always be true."""
        return True

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Anything":
        return cls()


class Less(_definitions.Comparator):
    """Allows anything less than the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Less than will be true."""
        return _operations.cast_compatible(self.value, observed) > observed


class LessOrEqual(_definitions.Comparator):
    """Allows anything less than or equal the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Less than or equal will be true."""
        return _operations.cast_compatible(self.value, observed) >= observed


class Greater(_definitions.Comparator):
    """Allows anything greater than the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Greater than will be true."""
        return _operations.cast_compatible(self.value, observed) < observed


class GreaterOrEqual(_definitions.Comparator):
    """Allows anything greater than or equal to the given value."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Greater than or equal will be true."""
        return _operations.cast_compatible(self.value, observed) <= observed


class Between(_definitions.Comparator):
    """Allows between the given values."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Greater than or equal will be true."""
        casted_min = _operations.cast_compatible(self.value["min"], observed)
        casted_max = _operations.cast_compatible(self.value["max"], observed)
        return casted_min <= observed <= casted_max

    @classmethod
    def construct(cls, minimum: typing.Any, maximum: typing.Any) -> "Between":
        return cls({"min": minimum, "max": maximum})

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Between":
        if isinstance(node, yaml.SequenceNode):
            loaded = loader.construct_sequence(node, deep=True)
            value = {"min": loaded[0], "max": loaded[1]}
        else:
            value = loader.construct_mapping(node, deep=True)
        return cls(value)


class OneOf(_definitions.Comparator):
    """Allows a matching comparison between any of the listed values."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Succeeds if at least one of the options are equal."""
        failures: typing.Dict[str, _definitions.Comparison] = {}
        for index, option in enumerate(self.value["options"]):
            if isinstance(option, _definitions.Comparator):
                comparator = option
            else:
                comparator = Equals(option)

            result = comparator.compare(observed, subset=subset)
            if getattr(result, "success", result):
                return result

            if isinstance(result, _definitions.Comparison):
                failures[str(index)] = result
            else:
                failures[str(index)] = _definitions.Comparison(
                    operation=comparator.operation_name(),
                    success=False,
                    expected=comparator.value,
                    observed=observed,
                )

        return _definitions.Comparison(
            operation="one_of",
            success=False,
            expected=", ".join([f"({i}) {f.expected}" for i, f in failures.items()]),
            observed=observed,
        )

    @classmethod
    def construct(cls, options: typing.List[typing.Any]) -> "OneOf":
        return cls({"options": options})

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "OneOf":
        options = loader.construct_sequence(node, deep=True)
        return cls({"options": options})


class NoneOf(_definitions.Comparator):
    """Allows a mismatching comparison between none of the listed values."""

    def _compare(
        self,
        observed: typing.Any,
        subset: bool = False,
    ) -> typing.Union[_definitions.Comparison, bool]:
        """Succeeds if none of the options are equal."""
        for index, option in enumerate(self.value["options"]):
            if isinstance(option, _definitions.Comparator):
                comparator = option
            else:
                comparator = Equals(option)

            result = comparator.compare(observed, subset=subset)
            if getattr(result, "success", False):
                return _definitions.Comparison(
                    operation=f"not {result.operation}",
                    success=False,
                    expected=result.expected,
                    observed=result.observed,
                    children=result.children,
                )

        return _definitions.Comparison(
            operation="none_of",
            success=True,
            expected=self.value,
            observed=observed,
        )

    @classmethod
    def construct(cls, options: typing.List[typing.Any]) -> "NoneOf":
        return cls({"options": options})

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "NoneOf":
        options = loader.construct_sequence(node, deep=True)
        return cls({"options": options})


Anything.register()
anything = getattr(Anything, "constructor", Anything)

Between.register()
between = getattr(Between, "constructor", Between)

Equals.register()
equals = getattr(Equals, "constructor", Equals)

Unequals.register()
unequals = getattr(Unequals, "constructor", Unequals)

Greater.register()
greater = getattr(Greater, "constructor", Greater)

GreaterOrEqual.register()
greater_or_equal = getattr(GreaterOrEqual, "constructor", GreaterOrEqual)

Less.register()
less = getattr(Less, "constructor", Less)

LessOrEqual.register()
less_or_equal = getattr(LessOrEqual, "constructor", LessOrEqual)

NoneOf.register()
none_of = getattr(NoneOf, "constructor", NoneOf)

OneOf.register()
one_of = getattr(OneOf, "constructor", OneOf)



from aok.comparisons._basics import Anything  # noqa: F401
from aok.comparisons._basics import Between  # noqa: F401
from aok.comparisons._basics import Equals  # noqa: F401
from aok.comparisons._basics import Greater  # noqa: F401
from aok.comparisons._basics import GreaterOrEqual  # noqa: F401
from aok.comparisons._basics import Less  # noqa: F401
from aok.comparisons._basics import LessOrEqual  # noqa: F401
from aok.comparisons._basics import NoneOf  # noqa: F401
from aok.comparisons._basics import OneOf  # noqa: F401
from aok.comparisons._basics import Unequals  # noqa: F401
from aok.comparisons._basics import anything  # noqa: F401
from aok.comparisons._basics import between  # noqa: F401
from aok.comparisons._basics import equals  # noqa: F401
from aok.comparisons._basics import greater  # noqa: F401
from aok.comparisons._basics import greater_or_equal  # noqa: F401
from aok.comparisons._basics import less  # noqa: F401
from aok.comparisons._basics import less_or_equal  # noqa: F401
from aok.comparisons._basics import none_of  # noqa: F401
from aok.comparisons._basics import one_of  # noqa: F401
from aok.comparisons._basics import unequals  # noqa: F401
from aok.comparisons._dicts import Dict  # noqa: F401
from aok.comparisons._dicts import Okay  # noqa: F401
from aok.comparisons._lists import JsonList  # noqa: F401
from aok.comparisons._lists import List  # noqa: F401
from aok.comparisons._lists import OkayList  # noqa: F401
from aok.comparisons._lists import Tuple  # noqa: F401
from aok.comparisons._lists import json_list  # noqa: F401
from aok.comparisons._nullish import NotNull  # noqa: F401
from aok.comparisons._nullish import Optional  # noqa: F401
from aok.comparisons._nullish import not_null  # noqa: F401
from aok.comparisons._nullish import optional  # noqa: F401
from aok.comparisons._strings import Like  # noqa: F401
from aok.comparisons._strings import LikeCase  # noqa: F401
from aok.comparisons._strings import Match  # noqa: F401
from aok.comparisons._strings import like  # noqa: F401
from aok.comparisons._strings import like_case  # noqa: F401
from aok.comparisons._strings import match  # noqa: F401



import json
import textwrap
import typing

import yaml
import yaml.constructor

import aok
from aok import _definitions
from aok import _types


def _compare_list(
    expected: "_types.ArbitraryList",
    observed: typing.Any,
    subset: bool,
) -> "_definitions.Comparison":
    """Compares lists recursively and returns the results as a Comparison."""
    expected_value = expected or []
    observed_value = observed or []

    if not isinstance(observed_value, list):
        return _definitions.Comparison(
            operation="list_type",
            success=False,
            expected=str(type(expected_value)),
            observed=str(type(observed_value)),
        )

    if len(expected_value) != len(observed_value):
        return _definitions.Comparison(
            operation="list_length",
            success=False,
            expected=len(expected_value),
            observed=len(observed_value),
        )

    results: typing.List[_definitions.Comparison] = [
        aok.to_comparator(exp).compare(obs, subset)
        for exp, obs in zip(expected_value, observed_value)
    ]

    return _definitions.Comparison(
        operation="list_comparison",
        success=all([r.success for r in results]),
        expected=expected_value,
        observed=observed,
        children={f"index_{i}": r for i, r in enumerate(results)},
    )


class List(_definitions.Comparator):
    """Container class for list comparisons, which compare the lists element-wise."""

    def compare(
        self,
        observed: typing.List[typing.Any],
        subset: bool = False,
    ) -> _definitions.Comparison:
        return _compare_list(
            expected=typing.cast(_types.ArbitraryList, self.value),
            observed=observed,
            subset=subset,
        )


class JsonList(_definitions.Comparator):
    """List comparator for data stored as a JSON string."""

    def compare(
        self,
        observed: str,
        subset: bool = False,
    ) -> _definitions.Comparison:
        """Parses and compares the observed value."""
        try:
            observed_parsed = json.loads(observed or "{}")
            if not isinstance(observed_parsed, list):
                raise ValueError("Not a JSON-serialized list.")
        except Exception as error:
            return _definitions.Comparison(
                operation="json_dict",
                success=False,
                expected=self.value,
                observed=observed,
                error=error,
            )

        return _compare_list(
            expected=typing.cast(_types.ArbitraryList, self.value),
            observed=observed_parsed,
            subset=subset,
        )

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "JsonList":
        """Loads the list from a yaml parser."""
        try:
            return cls(value=loader.construct_sequence(node, deep=True))
        except yaml.constructor.ConstructorError:
            if node.value == "":
                return cls(value={})
            raise


class Tuple(_definitions.Comparator):
    """Container class for tuple comparisons, which compare the tuples element-wise."""

    def compare(
        self,
        observed: typing.Tuple[typing.Any, ...],
        subset: bool = False,
    ) -> _definitions.Comparison:
        value = typing.cast(typing.Tuple[typing.Any, ...], self.value or tuple())
        other = observed or tuple()

        if len(value) != len(other):
            return _definitions.Comparison(
                operation="tuple_length",
                success=False,
                expected=len(value),
                observed=len(other),
            )

        results: typing.List[_definitions.Comparison] = [
            aok.to_comparator(exp).compare(obs, subset)
            for exp, obs in zip(value, other)
        ]

        return _definitions.Comparison(
            operation=self.operation_name(),
            success=all([r.success for r in results]),
            expected=self.value,
            observed=observed,
            children={f"index_{i}": r for i, r in enumerate(results)},
        )


class OkayList(List):
    """Root list object for comparison."""

    def assert_subset(self, observed: "_types.ArbitraryList", message: str = None):
        """
        Compares the observed object against the expected values in a recursive,
        element-wise fashion and raises assertion errors for any deviation between
        the elements of the expected configuration and the observed structure values.

        In this subset mode, any extra keys/values found in dictionaries will be
        ignored and assumed to be insignificant. Use `assert_all` for exact matching.

        :param observed:
            Data structure to compare against the expected one.
        :param message:
            Optional assertion message to include when differences are found between
            the observed and expected data structures.
        """
        result = self.compare(observed, subset=True)
        heading = message or "One or more subset differences were found"
        assert result.success, "{}\n{}".format(
            heading,
            textwrap.indent(result.to_diff_info() or "", "  "),
        )

    def assert_all(self, observed: "_types.ArbitraryList", message: str = None):
        """
        Compares the observed object against the expected values in a recursive,
        element-wise fashion and raises assertion errors for any deviation between
        the elements of the expected configuration and the observed structure values.

        :param observed:
            Data structure to compare against the expected one.
        :param message:
            Optional assertion message to include when differences are found between
            the observed and expected data structures.
        """
        result = self.compare(observed, subset=False)
        heading = message or "One or more exact differences were found"
        assert result.success, "{}\n{}".format(
            heading,
            textwrap.indent(result.to_diff_info() or "", "  "),
        )

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "OkayList":
        """Loads the list from a yaml parser."""
        try:
            return cls(value=loader.construct_sequence(node, deep=True))
        except yaml.constructor.ConstructorError:
            if node.value == "":
                return cls(value={})
            raise

    @classmethod
    def register(cls):
        """Overrides the registration in this case for base registration."""
        yaml.add_constructor("!aok_list", cls.parse_yaml)


OkayList.register()
List.register()
Tuple.register()

JsonList.register()
json_list = getattr(JsonList, "constructor", JsonList)



import json
import textwrap
import typing

import yaml
import yaml.constructor

import aok
from aok import _definitions
from aok import _types


def _compare_dicts(
    expected: "_types.ArbitraryDict",
    observed: typing.Any,
    subset: bool,
) -> "_definitions.Comparison":
    """Compares dictionaries recursively and returns the results as a Comparison."""
    expected_value = expected or {}
    observed_value = observed or {}

    if not isinstance(observed_value, dict):
        return _definitions.Comparison(
            operation="dict_type",
            success=False,
            expected=str(type(expected_value)),
            observed=str(type(observed_value)),
        )

    keys = set(list(expected_value.keys()))
    if not subset:
        keys.update(set(list(observed_value.keys())))

    results: typing.Dict[typing.Any, _definitions.Comparison] = {
        key: (
            aok.to_comparator(expected_value.get(key)).compare(
                observed_value.get(key), subset
            )
        )
        for key in keys
    }

    return _definitions.Comparison(
        operation="dict_comparison",
        success=all([r.success for r in results.values()]),
        expected=expected_value,
        observed=observed,
        children=results,
    )


class Dict(_definitions.Comparator):
    """Main class in which aok assertions are made."""

    def compare(
        self,
        observed: "_types.ArbitraryDict",
        subset: bool = False,
    ) -> _definitions.Comparison:
        return _compare_dicts(
            expected=typing.cast(_types.ArbitraryDict, self.value or {}),
            observed=observed,
            subset=subset,
        )


class JsonDict(_definitions.Comparator):
    """Dictionary comparator for data stored as a JSON string."""

    def compare(
        self,
        observed: str,
        subset: bool = False,
    ) -> _definitions.Comparison:
        """Parses and compares the observed value."""
        try:
            observed_parsed = json.loads(observed or "{}")
            if not isinstance(observed_parsed, dict):
                raise ValueError("Not a JSON-serialized dictionary.")
        except Exception as error:
            return _definitions.Comparison(
                operation="json_dict",
                success=False,
                expected=self.value,
                observed=observed,
                error=error,
            )

        return _compare_dicts(
            expected=typing.cast(_types.ArbitraryDict, self.value or {}),
            observed=observed_parsed,
            subset=subset,
        )

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "JsonDict":
        """Loads the dict from a yaml parser."""
        try:
            return cls(value=loader.construct_mapping(node, deep=True))
        except yaml.constructor.ConstructorError:
            if node.value == "":
                return cls(value={})
            raise


class Okay(Dict):
    """Root dictionary object for comparison."""

    def assert_subset(
        self,
        observed: "_types.ArbitraryDict",
        message: str = None,
    ) -> None:
        """
        Compares the observed object against the expected values in a recursive,
        element-wise fashion and raises assertion errors for any deviation between
        the elements of the expected configuration and the observed structure values.

        In this subset mode, any extra keys/values found in dictionaries will be
        ignored and assumed to be insignificant. Use `assert_all` for exact matching.

        :param observed:
            Data structure to compare against the expected one.
        :param message:
            Optional assertion message to include when differences are found between
            the observed and expected data structures.
        """
        result = self.compare(observed, subset=True)
        heading = message or "One or more subset differences were found"
        assert result.success, "{}\n{}".format(
            heading,
            textwrap.indent(result.to_diff_info() or "", "  "),
        )

    def assert_all(
        self,
        observed: "_types.ArbitraryDict",
        message: str = None,
    ) -> None:
        """
        Compares the observed object against the expected values in a recursive,
        element-wise fashion and raises assertion errors for any deviation between
        the elements of the expected configuration and the observed structure values.

        :param observed:
            Data structure to compare against the expected one.
        :param message:
            Optional assertion message to include when differences are found between
            the observed and expected data structures.
        """
        result = self.compare(observed, subset=False)
        heading = message or "One or more exact differences were found"
        assert result.success, "{}\n{}".format(
            heading,
            textwrap.indent(result.to_diff_info() or "", "  "),
        )

    @classmethod
    def _from_yaml(cls, loader: yaml.Loader, node: yaml.Node) -> "Okay":
        """Loads the dict from a yaml parser."""
        try:
            return cls(value=loader.construct_mapping(node, deep=True))
        except yaml.constructor.ConstructorError:
            if node.value == "":
                return cls(value={})
            raise

    @classmethod
    def register(cls):
        """Overrides the registration in this case for base registration."""
        yaml.add_constructor("!aok", Okay.parse_yaml)


Okay.register()

JsonDict.register()
